@@ -2,7 +2,7 @@
  * This file is part of libeval, a simple math expression evaluator
  *
  * Copyright (C) 2017 Michael Geselbracht, mgeselbracht3@gmail.com
- * Copyright (C) 2019-2021 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright (C) 2019-2023 KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 #include <vector>
 #include <algorithm>
 
+#include <eda_units.h>
 #include <string_utils.h>
 #include <wx/log.h>
 
@@ -116,14 +117,14 @@ bool VALUE::EqualTo( CONTEXT* aCtx, const VALUE* b ) const
 
     if( m_type == VT_NUMERIC && b->m_type == VT_NUMERIC )
     {
-        return m_valueDbl == b->m_valueDbl;
+        return AsDouble() == b->AsDouble();
     }
     else if( m_type == VT_STRING && b->m_type == VT_STRING )
     {
         if( b->m_stringIsWildcard )
-            return WildCompareString( b->m_valueStr, m_valueStr, false );
+            return WildCompareString( b->AsString(), AsString(), false );
         else
-            return !m_valueStr.CmpNoCase( b->m_valueStr );
+            return AsString().IsSameAs( b->AsString(), false );
     }
 
     return false;
@@ -200,6 +201,24 @@ wxString UCODE::Dump() const
 };
 
 
+wxString TOKENIZER::GetString()
+{
+    wxString rv;
+
+    while( m_pos < m_str.length() && m_str[ m_pos ] != '\'' )
+    {
+        if( m_str[ m_pos ] == '\\' && m_pos + 1 < m_str.length() && m_str[ m_pos + 1 ] == '\'' )
+            m_pos++;
+
+        rv.append( 1, m_str[ m_pos++ ] );
+    }
+
+    m_pos++;
+
+    return rv;
+}
+
+
 wxString TOKENIZER::GetChars( const std::function<bool( wxUniChar )>& cond ) const
 {
     wxString rv;
@@ -249,6 +268,7 @@ COMPILER::~COMPILER()
     if( m_tree )
     {
         freeTree( m_tree );
+        m_tree = nullptr;
     }
 
     // Allow explicit call to destructor
@@ -354,10 +374,11 @@ void COMPILER::newString( const wxString& aString )
     m_parseFinished = false;
 }
 
+
 T_TOKEN COMPILER::getToken()
 {
     T_TOKEN rv;
-    rv.value.str = nullptr;
+    rv.value = defaultTokenValue;
 
     bool    done = false;
 
@@ -380,12 +401,9 @@ T_TOKEN COMPILER::getToken()
 
 bool COMPILER::lexString( T_TOKEN& aToken )
 {
-    wxString str = m_tokenizer.GetChars( []( int c ) -> bool { return c != '\''; } );
-
     aToken.token = G_STRING;
-    aToken.value.str = new wxString( str );
+    aToken.value.str = new wxString( m_tokenizer.GetString() );
 
-    m_tokenizer.NextChar( str.length() + 1 );
     m_lexerState = LS_DEFAULT;
     return true;
 }
@@ -579,7 +597,7 @@ bool COMPILER::lexDefault( T_TOKEN& aToken )
 
 const wxString formatNode( TREE_NODE* node )
 {
-    return node->value.str ? *(node->value.str) : "";
+    return node->value.str ? *(node->value.str) : wxString( wxEmptyString );
 }
 
 
@@ -643,7 +661,7 @@ void dumpNode( wxString& buf, TREE_NODE* tok, int depth = 0 )
 
      case TR_OP_FUNC_CALL:
         buf += "CALL '";
-        buf += *tok->leaf[0]->value.str;
+        buf += formatNode( tok->leaf[0] );
         buf += "': ";
         dumpNode( buf, tok->leaf[1], depth + 1 );
         break;
@@ -693,6 +711,7 @@ void COMPILER::freeTree( LIBEVAL::TREE_NODE *tree )
         freeTree( tree->leaf[1] );
 
     delete tree->uop;
+    tree->uop = nullptr;
 }
 
 
@@ -782,7 +801,7 @@ static std::vector<TREE_NODE*> squashParamList( TREE_NODE* root )
     std::reverse( args.begin(), args.end() );
 
     for( size_t i = 0; i < args.size(); i++ )
-        libeval_dbg( 10, "squash arg%d: %s\n", int( i ), *args[i]->value.str );
+        libeval_dbg( 10, "squash arg%d: %s\n", int( i ), formatNode( args[i] ) );
 
     return args;
 }
@@ -841,7 +860,7 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
                 int pos = node->leaf[0]->srcPos;
 
                 if( node->leaf[0]->value.str )
-                    pos -= static_cast<int>( node->leaf[0]->value.str->length() );
+                    pos -= static_cast<int>( formatNode( node->leaf[0] ).length() );
 
                 reportError( CST_CODEGEN,  _( "Unknown parent of property" ), pos );
 
@@ -860,8 +879,8 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
                     // leaf[0]: object
                     // leaf[1]: field
 
-                    wxString itemName = *node->leaf[0]->value.str;
-                    wxString propName = *node->leaf[1]->value.str;
+                    wxString itemName = formatNode( node->leaf[0] );
+                    wxString propName = formatNode( node->leaf[1] );
                     std::unique_ptr<VAR_REF> vref = aCode->CreateVarRef( itemName, propName );
 
                     if( !vref )
@@ -891,7 +910,7 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
                     //    leaf[0]: function name
                     //    leaf[1]: parameter
 
-                    wxString                 itemName = *node->leaf[0]->value.str;
+                    wxString                 itemName = formatNode( node->leaf[0] );
                     std::unique_ptr<VAR_REF> vref = aCode->CreateVarRef( itemName, "" );
 
                     if( !vref )
@@ -901,7 +920,7 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
                                      node->leaf[0]->srcPos - (int) itemName.length() );
                     }
 
-                    wxString functionName = *node->leaf[1]->leaf[0]->value.str;
+                    wxString functionName = formatNode( node->leaf[1]->leaf[0] );
                     auto  func = aCode->CreateFuncCall( functionName );
                     std::vector<TREE_NODE*> params = squashParamList( node->leaf[1]->leaf[1] );
 
@@ -920,7 +939,7 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
                         for( TREE_NODE* pnode : params )
                         {
                             VALUE*   param = aPreflightContext->AllocValue();
-                            param->Set( *pnode->value.str );
+                            param->Set( formatNode( pnode ) );
                             aPreflightContext->Push( param );
                         }
 
@@ -964,8 +983,8 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
                     // leaf[0]: object
                     // leaf[1]: malformed syntax
 
-                    wxString itemName = *node->leaf[0]->value.str;
-                    wxString propName = *node->leaf[1]->value.str;
+                    wxString itemName = formatNode( node->leaf[0] );
+                    wxString propName = formatNode( node->leaf[1] );
                     std::unique_ptr<VAR_REF> vref = aCode->CreateVarRef( itemName, propName );
 
                     if( !vref )
@@ -1000,18 +1019,27 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
             }
             else if( son && son->op == TR_UNIT )
             {
+                if( m_unitResolver->GetSupportedUnits().empty() )
+                {
+                    msg.Printf( _( "Unexpected units for '%s'" ), formatNode( node ) );
+                    reportError( CST_CODEGEN, msg, node->srcPos );
+                }
+
                 int units = son->value.idx;
-                value =  m_unitResolver->Convert( *node->value.str, units );
+                value =  m_unitResolver->Convert( formatNode( node ), units );
                 son->isVisited = true;
             }
             else
             {
-                msg.Printf( _( "Missing units for '%s'| (%s)" ),
-                            *node->value.str,
-                            m_unitResolver->GetSupportedUnitsMessage() );
-                reportError( CST_CODEGEN, msg, node->srcPos );
+                if( !m_unitResolver->GetSupportedUnitsMessage().empty() )
+                {
+                    msg.Printf( _( "Missing units for '%s'| (%s)" ),
+                                formatNode( node ),
+                                m_unitResolver->GetSupportedUnitsMessage() );
+                    reportError( CST_CODEGEN, msg, node->srcPos );
+                }
 
-                value = DoubleValueFromString( EDA_UNITS::UNSCALED, *node->value.str );
+                value = EDA_UNIT_UTILS::UI::DoubleValueFromString( formatNode( node ) );
             }
 
             node->SetUop( TR_UOP_PUSH_VALUE, value );
@@ -1021,7 +1049,7 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
 
         case TR_STRING:
         {
-            wxString str = *node->value.str;
+            wxString str = formatNode( node );
             bool isWildcard = str.Contains("?") || str.Contains("*");
             node->SetUop( TR_UOP_PUSH_VALUE, str, isWildcard );
             node->isTerminal = true;
@@ -1030,12 +1058,12 @@ bool COMPILER::generateUCode( UCODE* aCode, CONTEXT* aPreflightContext )
 
         case TR_IDENTIFIER:
         {
-            std::unique_ptr<VAR_REF> vref = aCode->CreateVarRef( *node->value.str, "" );
+            std::unique_ptr<VAR_REF> vref = aCode->CreateVarRef( formatNode( node ), "" );
 
             if( !vref )
             {
-                msg.Printf( _( "Unrecognized item '%s'" ), *node->value.str );
-                reportError( CST_CODEGEN, msg, node->srcPos - (int) node->value.str->length() );
+                msg.Printf( _( "Unrecognized item '%s'" ), formatNode( node ) );
+                reportError( CST_CODEGEN, msg, node->srcPos - (int) formatNode( node ).length() );
             }
 
             node->SetUop( TR_UOP_PUSH_VAR, std::move( vref ) );
@@ -1113,12 +1141,12 @@ void UOP::Exec( CONTEXT* ctx )
         break;
     }
 
+#define AS_DOUBLE( arg ) ( arg ? arg->AsDouble() : 0.0 )
+
     if( m_op & TR_OP_BINARY_MASK )
     {
         LIBEVAL::VALUE* arg2 = ctx->Pop();
         LIBEVAL::VALUE* arg1 = ctx->Pop();
-        double          arg2Value = arg2 ? arg2->AsDouble() : 0.0;
-        double          arg1Value = arg1 ? arg1->AsDouble() : 0.0;
         double          result;
 
         if( ctx->HasErrorCallback() )
@@ -1140,28 +1168,28 @@ void UOP::Exec( CONTEXT* ctx )
         switch( m_op )
         {
         case TR_OP_ADD:
-            result = arg1Value + arg2Value;
+            result = AS_DOUBLE( arg1 ) + AS_DOUBLE( arg2 );
             break;
         case TR_OP_SUB:
-            result = arg1Value - arg2Value;
+            result = AS_DOUBLE( arg1 ) - AS_DOUBLE( arg2 );
             break;
         case TR_OP_MUL:
-            result = arg1Value * arg2Value;
+            result = AS_DOUBLE( arg1 ) * AS_DOUBLE( arg2 );
             break;
         case TR_OP_DIV:
-            result = arg1Value / arg2Value;
+            result = AS_DOUBLE( arg1 ) / AS_DOUBLE( arg2 );
             break;
         case TR_OP_LESS_EQUAL:
-            result = arg1Value <= arg2Value ? 1 : 0;
+            result = AS_DOUBLE( arg1 ) <= AS_DOUBLE( arg2 ) ? 1 : 0;
             break;
         case TR_OP_GREATER_EQUAL:
-            result = arg1Value >= arg2Value ? 1 : 0;
+            result = AS_DOUBLE( arg1 ) >= AS_DOUBLE( arg2 ) ? 1 : 0;
             break;
         case TR_OP_LESS:
-            result = arg1Value < arg2Value ? 1 : 0;
+            result = AS_DOUBLE( arg1 ) < AS_DOUBLE( arg2 ) ? 1 : 0;
             break;
         case TR_OP_GREATER:
-            result = arg1Value > arg2Value ? 1 : 0;
+            result = AS_DOUBLE( arg1 ) > AS_DOUBLE( arg2 ) ? 1 : 0;
             break;
         case TR_OP_EQUAL:
             if( !arg1 || !arg2 )
@@ -1180,10 +1208,10 @@ void UOP::Exec( CONTEXT* ctx )
                 result = arg1->NotEqualTo( ctx, arg2 ) ? 1 : 0;
             break;
         case TR_OP_BOOL_AND:
-            result = arg1Value != 0.0 && arg2Value != 0.0 ? 1 : 0;
+            result = AS_DOUBLE( arg1 ) != 0.0 && AS_DOUBLE( arg2 ) != 0.0 ? 1 : 0;
             break;
         case TR_OP_BOOL_OR:
-            result = arg1Value != 0.0 || arg2Value != 0.0 ? 1 : 0;
+            result = AS_DOUBLE( arg1 ) != 0.0 || AS_DOUBLE( arg2 ) != 0.0 ? 1 : 0;
             break;
         default:
             result = 0.0;
@@ -1198,16 +1226,16 @@ void UOP::Exec( CONTEXT* ctx )
     else if( m_op & TR_OP_UNARY_MASK )
     {
         LIBEVAL::VALUE* arg1 = ctx->Pop();
-        double          arg1Value = arg1 ? arg1->AsDouble() : 0.0;
+        double          ARG1VALUE = arg1 ? arg1->AsDouble() : 0.0;
         double          result;
 
         switch( m_op )
         {
         case TR_OP_BOOL_NOT:
-            result = arg1Value != 0.0 ? 0 : 1;
+            result = ARG1VALUE != 0.0 ? 0 : 1;
             break;
         default:
-            result = 0.0;
+            result = ARG1VALUE != 0.0 ? 1 : 0;
             break;
         }
 
