@@ -24,15 +24,16 @@
 
 #include <kiplatform/environment.h>
 #include <paths.h>
-#include <pgm_base.h>
-#include <settings/settings_manager.h>
 #include <config.h>
+#include <build_version.h>
+#include <macros.h>
+#include <wx_filename.h>
 
 // lowercase or pretty case depending on platform
 #if defined( __WXMAC__ ) || defined( __WXMSW__ )
 #define KICAD_PATH_STR wxT( "KiCad" )
 #else
-#define KICAD_PATH_STR wxT( "kicad" )
+#define KICAD_PATH_STR  wxT( "kicad" )
 #endif
 
 
@@ -46,7 +47,7 @@ void PATHS::getUserDocumentPath( wxFileName& aPath )
         aPath.AssignDir( KIPLATFORM::ENV::GetDocumentsPath() );
 
     aPath.AppendDir( KICAD_PATH_STR );
-    aPath.AppendDir( SETTINGS_MANAGER::GetSettingsVersion() );
+    aPath.AppendDir( GetMajorMinorVersion().ToStdString() );
 }
 
 
@@ -165,9 +166,13 @@ wxString PATHS::GetStockDataPath( bool aRespectRunFromBuildDir )
 #elif defined( __WXMSW__ )
         path = getWindowsKiCadRoot();
 #else
-        path = Pgm().GetExecutablePath() + wxT( ".." );
+        path = GetExecutablePath() + wxT( ".." );
 #endif
     }
+    else if( wxGetEnv( wxT( "KICAD_STOCK_DATA_HOME" ), &path ) && !path.IsEmpty() )
+    {
+        return path;
+    }
     else
     {
 #if defined( __WXMAC__ )
@@ -183,6 +188,17 @@ wxString PATHS::GetStockDataPath( bool aRespectRunFromBuildDir )
 }
 
 
+#ifdef __WXMSW__
+/**
+ * Gets the stock (install) data path, which is the base path for things like scripting, etc
+ */
+wxString PATHS::GetWindowsBaseSharePath()
+{
+    return getWindowsKiCadRoot() + wxT( "share\\" );
+}
+#endif
+
+
 wxString PATHS::GetStockEDALibraryPath()
 {
     wxString path;
@@ -203,7 +219,7 @@ wxString PATHS::GetStockSymbolsPath()
 {
     wxString path;
 
-    path = GetStockDataPath() + wxT( "/symbols" );
+    path = GetStockEDALibraryPath() + wxT( "/symbols" );
 
     return path;
 }
@@ -213,7 +229,7 @@ wxString PATHS::GetStockFootprintsPath()
 {
     wxString path;
 
-    path = GetStockDataPath() + wxT( "/footprints" );
+    path = GetStockEDALibraryPath() + wxT( "/footprints" );
 
     return path;
 }
@@ -223,7 +239,7 @@ wxString PATHS::GetStock3dmodelsPath()
 {
     wxString path;
 
-    path = GetStockDataPath() + wxT( "/3dmodels" );
+    path = GetStockEDALibraryPath() + wxT( "/3dmodels" );
 
     return path;
 }
@@ -243,7 +259,7 @@ wxString PATHS::GetStockTemplatesPath()
 {
     wxString path;
 
-    path = GetStockDataPath() + wxT( "/template" );
+    path = GetStockEDALibraryPath() + wxT( "/template" );
 
     return path;
 }
@@ -264,7 +280,7 @@ wxString PATHS::GetStockPluginsPath()
     wxFileName fn;
 
 #if defined( __WXMSW__ )
-    fn.AssignDir( Pgm().GetExecutablePath() );
+    fn.AssignDir( GetExecutablePath() );
     fn.AppendDir( wxT( "scripting" ) );
 #else
     fn.AssignDir( PATHS::GetStockDataPath( false ) );
@@ -279,17 +295,43 @@ wxString PATHS::GetStockPlugins3DPath()
 {
     wxFileName fn;
 
-#ifdef __WXGTK__
-    // KICAD_PLUGINDIR = CMAKE_INSTALL_FULL_LIBDIR path is the absolute path
-    // corresponding to the install path used for constructing KICAD_USER_PLUGIN
-    wxString tfname = wxString::FromUTF8Unchecked( KICAD_PLUGINDIR );
-    fn.Assign( tfname, wxEmptyString );
-    fn.AppendDir( wxT( "kicad" ) );
+#if defined( __WXMSW__ )
+    if( wxGetEnv( wxT( "KICAD_RUN_FROM_BUILD_DIR" ), nullptr ) )
+    {
+        fn.AssignDir( getWindowsKiCadRoot() );
+    }
+    else
+    {
+        fn.AssignDir( GetExecutablePath() );
+    }
+
     fn.AppendDir( wxT( "plugins" ) );
 #elif defined( __WXMAC__ )
     fn.Assign( wxStandardPaths::Get().GetPluginsDir(), wxEmptyString );
+
+    // This must be mapped to main bundle for everything but kicad.app
+    const wxArrayString dirs = fn.GetDirs();
+
+    // Check if we are the main kicad binary.  in this case, the path will be
+    //     /path/to/bundlename.app/Contents/PlugIns
+    // If we are an aux binary, the path will be something like
+    //     /path/to/bundlename.app/Contents/Applications/<standalone>.app/Contents/PlugIns
+    if( dirs.GetCount() >= 6 &&
+        dirs[dirs.GetCount() - 4] == wxT( "Applications" ) &&
+        dirs[dirs.GetCount() - 6].Lower().EndsWith( wxT( "app" ) ) )
+    {
+        fn.RemoveLastDir();
+        fn.RemoveLastDir();
+        fn.RemoveLastDir();
+        fn.RemoveLastDir();
+        fn.AppendDir( wxT( "PlugIns" ) );
+    }
 #else
-    fn.AssignDir( Pgm().GetExecutablePath() );
+    // KICAD_PLUGINDIR = CMAKE_INSTALL_FULL_LIBDIR path is the absolute path
+    // corresponding to the install path used for constructing KICAD_USER_PLUGIN
+    wxString tfname = wxString::FromUTF8Unchecked( KICAD_PLUGINDIR );
+    fn.Assign( tfname, "" );
+    fn.AppendDir( wxT( "kicad" ) );
     fn.AppendDir( wxT( "plugins" ) );
 #endif
 
@@ -312,11 +354,20 @@ wxString PATHS::GetStockDemosPath()
 
 wxString PATHS::GetUserCachePath()
 {
+    wxString   envPath;
     wxFileName tmp;
 
     tmp.AssignDir( KIPLATFORM::ENV::GetUserCachePath() );
+
+    // Use KICAD_CACHE_HOME to allow the user to force a specific cache path.
+    if( wxGetEnv( wxT( "KICAD_CACHE_HOME" ), &envPath ) && !envPath.IsEmpty() )
+    {
+        // Override the assignment above with KICAD_CACHE_HOME
+        tmp.AssignDir( envPath );
+    }
+
     tmp.AppendDir( KICAD_PATH_STR );
-    tmp.AppendDir( SETTINGS_MANAGER::GetSettingsVersion() );
+    tmp.AppendDir( GetMajorMinorVersion().ToStdString() );
 
     return tmp.GetPathWithSep();
 }
@@ -338,10 +389,20 @@ wxString PATHS::GetDocumentationPath()
 }
 
 
+wxString PATHS::GetInstanceCheckerPath()
+{
+    wxFileName path;
+    path.AssignDir( wxStandardPaths::Get().GetTempDir() );
+    path.AppendDir( "org.kicad.kicad" );
+    path.AppendDir( "instances" );
+    return path.GetPathWithSep();
+}
+
+
 bool PATHS::EnsurePathExists( const wxString& aPath )
 {
     wxFileName path( aPath );
-    if( !path.Normalize() )
+    if( !path.MakeAbsolute() )
     {
         return false;
     }
@@ -360,8 +421,8 @@ bool PATHS::EnsurePathExists( const wxString& aPath )
 
 void PATHS::EnsureUserPathsExist()
 {
+    EnsurePathExists( GetUserCachePath() );
     EnsurePathExists( GetUserPluginsPath() );
-    EnsurePathExists( GetUserPlugins3DPath() );
     EnsurePathExists( GetUserScriptingPath() );
     EnsurePathExists( GetUserTemplatesPath() );
     EnsurePathExists( GetDefaultUserProjectsPath() );
@@ -380,9 +441,9 @@ wxString PATHS::GetOSXKicadUserDataDir()
     wxFileName udir( wxStandardPaths::Get().GetUserDataDir(), wxEmptyString );
 
     // Since appname is different if started via launcher or standalone binary
-    // map all to wxT( "kicad" ) here
+    // map all to "kicad" here
     udir.RemoveLastDir();
-    udir.AppendDir( wxT( "kicad" ) );
+    udir.AppendDir(  wxT( "kicad" ) );
 
     return udir.GetPath();
 }
@@ -424,12 +485,109 @@ wxString PATHS::GetOSXKicadDataDir()
 #endif
 
 
-#ifdef __WXWINDOWS__
+#ifdef __WXMSW__
+wxString PATHS::GetWindowsFontConfigDir()
+{
+    wxFileName fn;
+    fn.AssignDir( getWindowsKiCadRoot() );
+    fn.AppendDir( wxS( "etc" ) );
+    fn.AppendDir( wxS( "fonts" ) );
+
+    return fn.GetPathWithSep();
+}
+
+
 wxString PATHS::getWindowsKiCadRoot()
 {
-    wxFileName root( Pgm().GetExecutablePath() + wxT( "/../" ) );
-    root.Normalize();
+    wxFileName root( GetExecutablePath() +  wxT( "/../" ) );
+    root.MakeAbsolute();
 
     return root.GetPathWithSep();
 }
 #endif
+
+
+wxString PATHS::GetUserSettingsPath()
+{
+    static wxString user_settings_path;
+
+    if( user_settings_path.empty() )
+        user_settings_path = CalculateUserSettingsPath();
+
+    return user_settings_path;
+}
+
+
+wxString PATHS::CalculateUserSettingsPath( bool aIncludeVer, bool aUseEnv )
+{
+    wxFileName cfgpath;
+
+    // http://docs.wxwidgets.org/3.0/classwx_standard_paths.html#a7c7cf595d94d29147360d031647476b0
+
+    wxString envstr;
+    if( aUseEnv && wxGetEnv( wxT( "KICAD_CONFIG_HOME" ), &envstr ) && !envstr.IsEmpty() )
+    {
+        // Override the assignment above with KICAD_CONFIG_HOME
+        cfgpath.AssignDir( envstr );
+    }
+    else
+    {
+        cfgpath.AssignDir( KIPLATFORM::ENV::GetUserConfigPath() );
+
+        cfgpath.AppendDir( TO_STR( KICAD_CONFIG_DIR ) );
+    }
+
+    if( aIncludeVer )
+        cfgpath.AppendDir( GetMajorMinorVersion().ToStdString() );
+
+    return cfgpath.GetPath();
+}
+
+
+const wxString& PATHS::GetExecutablePath()
+{
+    static wxString exe_path;
+
+    if( exe_path.empty() )
+    {
+        wxString bin_dir = wxStandardPaths::Get().GetExecutablePath();
+
+#ifdef __WXMAC__
+        // On OSX GetExecutablePath() will always point to main
+        // bundle directory, e.g., /Applications/kicad.app/
+
+        wxFileName fn( bin_dir );
+        WX_FILENAME::ResolvePossibleSymlinks( fn );
+
+        if( fn.GetName() == wxT( "kicad" ) || fn.GetName() == wxT( "kicad-cli" ) )
+        {
+            // kicad launcher, so just remove the Contents/MacOS part
+            fn.RemoveLastDir();
+            fn.RemoveLastDir();
+        }
+        else
+        {
+            // standalone binaries live in Contents/Applications/<standalone>.app/Contents/MacOS
+            fn.RemoveLastDir();
+            fn.RemoveLastDir();
+            fn.RemoveLastDir();
+            fn.RemoveLastDir();
+            fn.RemoveLastDir();
+        }
+
+        bin_dir = fn.GetPath() + wxT( "/" );
+#else
+        // Use unix notation for paths. I am not sure this is a good idea,
+        // but it simplifies compatibility between Windows and Unices.
+        // However it is a potential problem in path handling under Windows.
+        bin_dir.Replace( WIN_STRING_DIR_SEP, UNIX_STRING_DIR_SEP );
+
+        // Remove file name form command line:
+        while( bin_dir.Last() != '/' && !bin_dir.IsEmpty() )
+            bin_dir.RemoveLast();
+#endif
+        exe_path = bin_dir;
+    }
+
+    return exe_path;
+}
