@@ -2,7 +2,7 @@
  * This program source code file is part of KiCad, a free EDA CAD application.
  *
  * Copyright (C) 2015 Jean-Pierre Charras, jp.charras at wanadoo.fr
- * Copyright (C) 1992-2021 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright (C) 1992-2024 KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -22,10 +22,13 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
+#include <base_units.h>
 #include <bitmaps.h>
+#include <string_utils.h>
 #include <core/mirror.h>
 #include <sch_painter.h>
-#include <plotters/plotter.h>
+#include <sch_plotter.h>
+#include <geometry/shape_segment.h>
 #include <sch_line.h>
 #include <sch_edit_frame.h>
 #include <settings/color_settings.h>
@@ -35,15 +38,15 @@
 #include <project/net_settings.h>
 #include <trigo.h>
 #include <board_item.h>
-#include <advanced_config.h>
 
-SCH_LINE::SCH_LINE( const wxPoint& pos, int layer ) :
+
+SCH_LINE::SCH_LINE( const VECTOR2I& pos, int layer ) :
     SCH_ITEM( nullptr, SCH_LINE_T )
 {
     m_start           = pos;
     m_end             = pos;
     m_stroke.SetWidth( 0 );
-    m_stroke.SetPlotStyle( PLOT_DASH_TYPE::DEFAULT );
+    m_stroke.SetLineStyle( LINE_STYLE::DEFAULT );
     m_stroke.SetColor( COLOR4D::UNSPECIFIED );
 
     switch( layer )
@@ -59,13 +62,13 @@ SCH_LINE::SCH_LINE( const wxPoint& pos, int layer ) :
         m_startIsDangling = m_endIsDangling = false;
 
     if( layer == LAYER_WIRE )
-        m_lastResolvedWidth = Mils2iu( DEFAULT_WIRE_WIDTH_MILS );
+        m_lastResolvedWidth = schIUScale.MilsToIU( DEFAULT_WIRE_WIDTH_MILS );
     else if( layer == LAYER_BUS )
-        m_lastResolvedWidth = Mils2iu( DEFAULT_BUS_WIDTH_MILS );
+        m_lastResolvedWidth = schIUScale.MilsToIU( DEFAULT_BUS_WIDTH_MILS );
     else
-        m_lastResolvedWidth = Mils2iu( DEFAULT_LINE_WIDTH_MILS );
+        m_lastResolvedWidth = schIUScale.MilsToIU( DEFAULT_LINE_WIDTH_MILS );
 
-    m_lastResolvedLineStyle = GetDefaultStyle();
+    m_lastResolvedLineStyle = LINE_STYLE::SOLID;
     m_lastResolvedColor = COLOR4D::UNSPECIFIED;
 }
 
@@ -82,6 +85,19 @@ SCH_LINE::SCH_LINE( const SCH_LINE& aLine ) :
     m_lastResolvedLineStyle = aLine.m_lastResolvedLineStyle;
     m_lastResolvedWidth = aLine.m_lastResolvedWidth;
     m_lastResolvedColor = aLine.m_lastResolvedColor;
+
+    m_operatingPoint = aLine.m_operatingPoint;
+}
+
+
+wxString SCH_LINE::GetFriendlyName() const
+{
+    switch( GetLayer() )
+    {
+    case LAYER_WIRE: return _( "Wire" );
+    case LAYER_BUS:  return _( "Bus" );
+    default:         return _( "Graphic Line" );
+    }
 }
 
 
@@ -91,73 +107,22 @@ EDA_ITEM* SCH_LINE::Clone() const
 }
 
 
-/*
- * Conversion between PLOT_DASH_TYPE values and style names displayed
- */
-const std::map<PLOT_DASH_TYPE, const char*> lineStyleNames{
-    { PLOT_DASH_TYPE::SOLID, "solid" },
-    { PLOT_DASH_TYPE::DASH, "dashed" },
-    { PLOT_DASH_TYPE::DASHDOT, "dash_dot" },
-    { PLOT_DASH_TYPE::DOT, "dotted" },
-};
-
-
-const char* SCH_LINE::GetLineStyleName( PLOT_DASH_TYPE aStyle )
+void SCH_LINE::Move( const VECTOR2I& aOffset )
 {
-    auto resultIt = lineStyleNames.find( aStyle );
-
-    //legacy behavior is to default to dash if there is no name
-    return resultIt == lineStyleNames.end() ? lineStyleNames.find( PLOT_DASH_TYPE::DASH )->second :
-                                              resultIt->second;
+    m_start += aOffset;
+    m_end += aOffset;
 }
 
 
-PLOT_DASH_TYPE SCH_LINE::GetLineStyleByName( const wxString& aStyleName )
+void SCH_LINE::MoveStart( const VECTOR2I& aOffset )
 {
-    PLOT_DASH_TYPE id = PLOT_DASH_TYPE::DEFAULT; // Default style id
-
-    //find the name by value
-    auto resultIt = std::find_if( lineStyleNames.begin(), lineStyleNames.end(),
-                                  [aStyleName]( const auto& it )
-                                  {
-                                      return it.second == aStyleName;
-                                  } );
-
-    if( resultIt != lineStyleNames.end() )
-        id = resultIt->first;
-
-    return id;
+    m_start += aOffset;
 }
 
 
-void SCH_LINE::Move( const wxPoint& aOffset )
+void SCH_LINE::MoveEnd( const VECTOR2I& aOffset )
 {
-    if( aOffset != wxPoint( 0, 0 ) )
-    {
-        m_start += aOffset;
-        m_end += aOffset;
-        SetModified();
-    }
-}
-
-
-void SCH_LINE::MoveStart( const wxPoint& aOffset )
-{
-    if( aOffset != wxPoint( 0, 0 ) )
-    {
-        m_start += aOffset;
-        SetModified();
-    }
-}
-
-
-void SCH_LINE::MoveEnd( const wxPoint& aOffset )
-{
-    if( aOffset != wxPoint( 0, 0 ) )
-    {
-        m_end += aOffset;
-        SetModified();
-    }
+    m_end += aOffset;
 }
 
 
@@ -180,25 +145,50 @@ void SCH_LINE::Show( int nestLevel, std::ostream& os ) const
 
 void SCH_LINE::ViewGetLayers( int aLayers[], int& aCount ) const
 {
-    aCount     = 3;
+    aCount     = 4;
     aLayers[0] = LAYER_DANGLING;
     aLayers[1] = m_layer;
     aLayers[2] = LAYER_SELECTION_SHADOWS;
+    aLayers[3] = LAYER_OP_VOLTAGES;
 }
 
 
-const EDA_RECT SCH_LINE::GetBoundingBox() const
+double SCH_LINE::ViewGetLOD( int aLayer, KIGFX::VIEW* aView ) const
 {
-    int      width = m_stroke.GetWidth() / 2;
-    int      extra = m_stroke.GetWidth() & 0x1;
+    constexpr double HIDE = std::numeric_limits<double>::max();
+    constexpr double SHOW = 0.0;
 
-    int      xmin = std::min( m_start.x, m_end.x ) - width;
-    int      ymin = std::min( m_start.y, m_end.y ) - width;
+    if( aLayer == LAYER_OP_VOLTAGES )
+    {
+        if( m_start == m_end )
+            return HIDE;
 
-    int      xmax = std::max( m_start.x, m_end.x ) + width + extra;
-    int      ymax = std::max( m_start.y, m_end.y ) + width + extra;
+        int height = std::abs( m_end.y - m_start.y );
+        int width = std::abs( m_end.x - m_start.x );
 
-    EDA_RECT ret( wxPoint( xmin, ymin ), wxSize( xmax - xmin, ymax - ymin ) );
+        // Operating points will be shown only if zoom is appropriate
+        if( height == 0 )
+            return (double) schIUScale.mmToIU( 15 ) / width;
+        else
+            return (double) schIUScale.mmToIU( 5 ) / height;
+    }
+
+    // Other layers are always drawn.
+    return SHOW;
+}
+
+
+const BOX2I SCH_LINE::GetBoundingBox() const
+{
+    int   width = GetPenWidth() / 2;
+
+    int   xmin = std::min( m_start.x, m_end.x ) - width;
+    int   ymin = std::min( m_start.y, m_end.y ) - width;
+
+    int   xmax = std::max( m_start.x, m_end.x ) + width + 1;
+    int   ymax = std::max( m_start.y, m_end.y ) + width + 1;
+
+    BOX2I ret( VECTOR2I( xmin, ymin ), VECTOR2I( xmax - xmin, ymax - ymin ) );
 
     return ret;
 }
@@ -213,6 +203,7 @@ double SCH_LINE::GetLength() const
 void SCH_LINE::SetLineColor( const COLOR4D& aColor )
 {
     m_stroke.SetColor( aColor );
+    m_lastResolvedColor = GetLineColor();
 }
 
 
@@ -236,67 +227,46 @@ void SCH_LINE::SetLineColor( const double r, const double g, const double b, con
 COLOR4D SCH_LINE::GetLineColor() const
 {
     if( m_stroke.GetColor() != COLOR4D::UNSPECIFIED )
-    {
         m_lastResolvedColor = m_stroke.GetColor();
-    }
-    else if( IsConnectable() && !IsConnectivityDirty() )
-    {
-        NETCLASSPTR netclass = NetClass();
-
-        if( netclass )
-            m_lastResolvedColor = netclass->GetSchematicColor();
-    }
+    else if( !IsConnectable() )
+        m_lastResolvedColor = COLOR4D::UNSPECIFIED;
+    else if( !IsConnectivityDirty() )
+        m_lastResolvedColor = GetEffectiveNetClass()->GetSchematicColor();
 
     return m_lastResolvedColor;
 }
 
 
-PLOT_DASH_TYPE SCH_LINE::GetDefaultStyle() const
-{
-    if( IsGraphicLine() )
-        return PLOT_DASH_TYPE::DASH;
-
-    return PLOT_DASH_TYPE::SOLID;
-}
-
-
 void SCH_LINE::SetLineStyle( const int aStyleId )
 {
-    SetLineStyle( static_cast<PLOT_DASH_TYPE>( aStyleId ) );
+    SetLineStyle( static_cast<LINE_STYLE>( aStyleId ) );
 }
 
 
-void SCH_LINE::SetLineStyle( const PLOT_DASH_TYPE aStyle )
+void SCH_LINE::SetLineStyle( const LINE_STYLE aStyle )
 {
-    if( aStyle == GetDefaultStyle() )
-        m_stroke.SetPlotStyle( PLOT_DASH_TYPE::DEFAULT );
-    else
-        m_stroke.SetPlotStyle( aStyle );
+    m_stroke.SetLineStyle( aStyle );
+    m_lastResolvedLineStyle = GetLineStyle();
 }
 
 
-PLOT_DASH_TYPE SCH_LINE::GetLineStyle() const
+LINE_STYLE SCH_LINE::GetLineStyle() const
 {
-    if( m_stroke.GetPlotStyle() != PLOT_DASH_TYPE::DEFAULT )
-        return m_stroke.GetPlotStyle();
+    if( m_stroke.GetLineStyle() != LINE_STYLE::DEFAULT )
+        return m_stroke.GetLineStyle();
 
-    return GetDefaultStyle();
+    return LINE_STYLE::SOLID;
 }
 
 
-PLOT_DASH_TYPE SCH_LINE::GetEffectiveLineStyle() const
+LINE_STYLE SCH_LINE::GetEffectiveLineStyle() const
 {
-    if( m_stroke.GetPlotStyle() != PLOT_DASH_TYPE::DEFAULT )
-    {
-        m_lastResolvedLineStyle = m_stroke.GetPlotStyle();
-    }
-    else if( IsConnectable() && !IsConnectivityDirty() )
-    {
-        NETCLASSPTR netclass = NetClass();
-
-        if( netclass )
-            m_lastResolvedLineStyle = static_cast<PLOT_DASH_TYPE>( netclass->GetLineStyle() );
-    }
+    if( m_stroke.GetLineStyle() != LINE_STYLE::DEFAULT )
+        m_lastResolvedLineStyle = m_stroke.GetLineStyle();
+    else if( !IsConnectable() )
+        m_lastResolvedLineStyle = LINE_STYLE::SOLID;
+    else if( !IsConnectivityDirty() )
+        m_lastResolvedLineStyle = (LINE_STYLE) GetEffectiveNetClass()->GetLineStyle();
 
     return m_lastResolvedLineStyle;
 }
@@ -305,13 +275,13 @@ PLOT_DASH_TYPE SCH_LINE::GetEffectiveLineStyle() const
 void SCH_LINE::SetLineWidth( const int aSize )
 {
     m_stroke.SetWidth( aSize );
+    m_lastResolvedWidth = GetPenWidth();
 }
 
 
 int SCH_LINE::GetPenWidth() const
 {
     SCHEMATIC*  schematic = Schematic();
-    NETCLASSPTR netclass;
 
     switch ( m_layer )
     {
@@ -322,48 +292,28 @@ int SCH_LINE::GetPenWidth() const
         if( schematic )
             return schematic->Settings().m_DefaultLineWidth;
 
-        return Mils2iu( DEFAULT_LINE_WIDTH_MILS );
+        return schIUScale.MilsToIU( DEFAULT_LINE_WIDTH_MILS );
 
     case LAYER_WIRE:
         if( m_stroke.GetWidth() > 0 )
-        {
             m_lastResolvedWidth = m_stroke.GetWidth();
-        }
         else if( !IsConnectivityDirty() )
-        {
-            netclass = NetClass();
-
-            if( !netclass && schematic  )
-                netclass = schematic->Prj().GetProjectFile().NetSettings().m_NetClasses.GetDefault();
-
-            if( netclass )
-                m_lastResolvedWidth = netclass->GetWireWidth();
-        }
+            m_lastResolvedWidth = GetEffectiveNetClass()->GetWireWidth();
 
         return m_lastResolvedWidth;
 
     case LAYER_BUS:
         if( m_stroke.GetWidth() > 0 )
-        {
             m_lastResolvedWidth = m_stroke.GetWidth();
-        }
         else if( !IsConnectivityDirty() )
-        {
-            netclass = NetClass();
-
-            if( !netclass && schematic )
-                netclass = schematic->Prj().GetProjectFile().NetSettings().m_NetClasses.GetDefault();
-
-            if( netclass )
-                m_lastResolvedWidth = netclass->GetBusWidth();
-        }
+            m_lastResolvedWidth = GetEffectiveNetClass()->GetBusWidth();
 
         return m_lastResolvedWidth;
     }
 }
 
 
-void SCH_LINE::Print( const RENDER_SETTINGS* aSettings, const wxPoint& offset )
+void SCH_LINE::Print( const RENDER_SETTINGS* aSettings, const VECTOR2I& offset )
 {
     wxDC*   DC = aSettings->GetPrintDC();
     COLOR4D color = GetLineColor();
@@ -371,56 +321,24 @@ void SCH_LINE::Print( const RENDER_SETTINGS* aSettings, const wxPoint& offset )
     if( color == COLOR4D::UNSPECIFIED )
         color = aSettings->GetLayerColor( GetLayer() );
 
-    wxPoint        start = m_start;
-    wxPoint        end = m_end;
-    PLOT_DASH_TYPE lineStyle = GetEffectiveLineStyle();
-    int            penWidth = std::max( GetPenWidth(), aSettings->GetDefaultPenWidth() );
+    VECTOR2I   start = m_start;
+    VECTOR2I   end = m_end;
+    LINE_STYLE lineStyle = GetEffectiveLineStyle();
+    int        penWidth = std::max( GetPenWidth(), aSettings->GetDefaultPenWidth() );
 
-    if( lineStyle <= PLOT_DASH_TYPE::FIRST_TYPE )
+    if( lineStyle <= LINE_STYLE::FIRST_TYPE )
     {
-        GRLine( nullptr, DC, start.x, start.y, end.x, end.y, penWidth, color );
+        GRLine( DC, start.x, start.y, end.x, end.y, penWidth, color );
     }
     else
     {
-        EDA_RECT clip( (wxPoint) start, wxSize( end.x - start.x, end.y - start.y ) );
-        clip.Normalize();
+        SHAPE_SEGMENT segment( start, end );
 
-        double theta = atan2( end.y - start.y, end.x - start.x );
-        double strokes[] = { 1.0, dash_gap_len( penWidth ), 1.0, dash_gap_len( penWidth ) };
-
-        switch( lineStyle )
-        {
-        default:
-        case PLOT_DASH_TYPE::DASH:
-            strokes[0] = strokes[2] = dash_mark_len( penWidth );
-            break;
-        case PLOT_DASH_TYPE::DOT:
-            strokes[0] = strokes[2] = dot_mark_len( penWidth );
-            break;
-        case PLOT_DASH_TYPE::DASHDOT:
-            strokes[0] = dash_mark_len( penWidth );
-            strokes[2] = dot_mark_len( penWidth );
-            break;
-        }
-
-        for( size_t i = 0; i < 10000; ++i )
-        {
-            // Calculations MUST be done in doubles to keep from accumulating rounding
-            // errors as we go.
-            wxPoint next( start.x + strokes[ i % 4 ] * cos( theta ),
-                          start.y + strokes[ i % 4 ] * sin( theta ) );
-
-            // Drawing each segment can be done rounded to ints.
-            wxPoint segStart( KiROUND( start.x ), KiROUND( start.y ) );
-            wxPoint segEnd( KiROUND( next.x ), KiROUND( next.y ) );
-
-            if( ClipLine( &clip, segStart.x, segStart.y, segEnd.x, segEnd.y ) )
-                break;
-            else if( i % 2 == 0 )
-                GRLine( nullptr, DC, segStart.x, segStart.y, segEnd.x, segEnd.y, penWidth, color );
-
-            start = next;
-        }
+        STROKE_PARAMS::Stroke( &segment, lineStyle, penWidth, aSettings,
+                               [&]( const VECTOR2I& a, const VECTOR2I& b )
+                               {
+                                   GRLine( DC, a.x, a.y, b.x, b.y, penWidth, color );
+                               } );
     }
 }
 
@@ -445,53 +363,54 @@ void SCH_LINE::MirrorHorizontally( int aCenter )
 }
 
 
-void SCH_LINE::Rotate( const wxPoint& aCenter )
+void SCH_LINE::Rotate( const VECTOR2I& aCenter )
 {
     // When we allow off grid items, the
     // else if should become a plain if to allow
     // rotation around the center of the line
     if( m_flags & STARTPOINT )
-        RotatePoint( &m_start, aCenter, 900 );
+        RotatePoint( m_start, aCenter, ANGLE_90 );
+
     else if( m_flags & ENDPOINT )
-        RotatePoint( &m_end, aCenter, 900 );
+        RotatePoint( m_end, aCenter, ANGLE_90 );
 }
 
 
-void SCH_LINE::RotateStart( const wxPoint& aCenter )
+void SCH_LINE::RotateStart( const VECTOR2I& aCenter )
 {
-    RotatePoint( &m_start, aCenter, 900 );
+    RotatePoint( m_start, aCenter, ANGLE_90 );
 }
 
 
-void SCH_LINE::RotateEnd( const wxPoint& aCenter )
+void SCH_LINE::RotateEnd( const VECTOR2I& aCenter )
 {
-    RotatePoint( &m_end, aCenter, 900 );
+    RotatePoint( m_end, aCenter, ANGLE_90 );
 }
 
 
-int SCH_LINE::GetAngleFrom( const wxPoint& aPoint ) const
+int SCH_LINE::GetAngleFrom( const VECTOR2I& aPoint ) const
 {
-    wxPoint vec;
+    VECTOR2I vec;
 
     if( aPoint == m_start )
         vec = m_end - aPoint;
     else
         vec = m_start - aPoint;
 
-    return KiROUND( ArcTangente( vec.y, vec.x ) );
+    return KiROUND( EDA_ANGLE( vec ).AsDegrees() );
 }
 
 
-int SCH_LINE::GetReverseAngleFrom( const wxPoint& aPoint ) const
+int SCH_LINE::GetReverseAngleFrom( const VECTOR2I& aPoint ) const
 {
-    wxPoint vec;
+    VECTOR2I vec;
 
     if( aPoint == m_end )
         vec = m_start - aPoint;
     else
         vec = m_end - aPoint;
 
-    return KiROUND( ArcTangente( vec.y, vec.x ) );
+    return KiROUND( EDA_ANGLE( vec ).AsDegrees() );
 }
 
 
@@ -500,8 +419,8 @@ bool SCH_LINE::IsParallel( const SCH_LINE* aLine ) const
     wxCHECK_MSG( aLine != nullptr && aLine->Type() == SCH_LINE_T, false,
                  wxT( "Cannot test line segment for overlap." ) );
 
-    wxPoint firstSeg   = m_end - m_start;
-    wxPoint secondSeg = aLine->m_end - aLine->m_start;
+    VECTOR2I firstSeg = m_end - m_start;
+    VECTOR2I secondSeg = aLine->m_end - aLine->m_start;
 
     // Use long long here to avoid overflow in calculations
     return !( (long long) firstSeg.x * secondSeg.y - (long long) firstSeg.y * secondSeg.x );
@@ -511,7 +430,7 @@ bool SCH_LINE::IsParallel( const SCH_LINE* aLine ) const
 SCH_LINE* SCH_LINE::MergeOverlap( SCH_SCREEN* aScreen, SCH_LINE* aLine, bool aCheckJunctions )
 {
     auto less =
-            []( const wxPoint& lhs, const wxPoint& rhs ) -> bool
+            []( const VECTOR2I& lhs, const VECTOR2I& rhs ) -> bool
             {
                 if( lhs.x == rhs.x )
                     return lhs.y < rhs.y;
@@ -525,11 +444,11 @@ SCH_LINE* SCH_LINE::MergeOverlap( SCH_SCREEN* aScreen, SCH_LINE* aLine, bool aCh
     if( this == aLine || GetLayer() != aLine->GetLayer() )
         return nullptr;
 
-    wxPoint leftmost_start = aLine->m_start;
-    wxPoint leftmost_end = aLine->m_end;
+    VECTOR2I leftmost_start = aLine->m_start;
+    VECTOR2I leftmost_end = aLine->m_end;
 
-    wxPoint rightmost_start = m_start;
-    wxPoint rightmost_end = m_end;
+    VECTOR2I rightmost_start = m_start;
+    VECTOR2I rightmost_end = m_end;
 
     // We place the start to the left and below the end of both lines
     if( leftmost_start != std::min( { leftmost_start, leftmost_end }, less ) )
@@ -547,8 +466,8 @@ SCH_LINE* SCH_LINE::MergeOverlap( SCH_SCREEN* aScreen, SCH_LINE* aLine, bool aCh
         std::swap( leftmost_end, rightmost_end );
     }
 
-    wxPoint other_start = rightmost_start;
-    wxPoint other_end = rightmost_end;
+    VECTOR2I other_start = rightmost_start;
+    VECTOR2I other_end = rightmost_end;
 
     if( less( rightmost_end, leftmost_end ) )
     {
@@ -629,6 +548,18 @@ SCH_LINE* SCH_LINE::MergeOverlap( SCH_SCREEN* aScreen, SCH_LINE* aLine, bool aCh
 }
 
 
+SCH_LINE* SCH_LINE::BreakAt( const VECTOR2I& aPoint )
+{
+    SCH_LINE* newSegment = static_cast<SCH_LINE*>( Duplicate() );
+
+    newSegment->SetStartPoint( aPoint );
+    newSegment->SetConnectivityDirty( true );
+    SetEndPoint( aPoint );
+
+    return newSegment;
+}
+
+
 void SCH_LINE::GetEndPoints( std::vector <DANGLING_END_ITEM>& aItemList )
 {
     if( IsConnectable() )
@@ -639,44 +570,56 @@ void SCH_LINE::GetEndPoints( std::vector <DANGLING_END_ITEM>& aItemList )
 }
 
 
-bool SCH_LINE::UpdateDanglingState( std::vector<DANGLING_END_ITEM>& aItemList,
-                                    const SCH_SHEET_PATH* aPath )
+bool SCH_LINE::UpdateDanglingState( std::vector<DANGLING_END_ITEM>& aItemListByType,
+                                    std::vector<DANGLING_END_ITEM>& aItemListByPos,
+                                    const SCH_SHEET_PATH*           aPath )
 {
-    if( IsConnectable() )
+    if( !IsConnectable() )
+        return false;
+
+    bool previousStartState = m_startIsDangling;
+    bool previousEndState = m_endIsDangling;
+
+    m_startIsDangling = m_endIsDangling = true;
+
+    for( auto it = DANGLING_END_ITEM_HELPER::get_lower_pos( aItemListByPos, m_start );
+         it < aItemListByPos.end() && it->GetPosition() == m_start; it++ )
     {
-        bool previousStartState = m_startIsDangling;
-        bool previousEndState = m_endIsDangling;
+        DANGLING_END_ITEM& item = *it;
 
-        m_startIsDangling = m_endIsDangling = true;
+        if( item.GetItem() == this )
+            continue;
 
-        for( DANGLING_END_ITEM item : aItemList )
+        if( ( IsWire() && item.GetType() != BUS_END && item.GetType() != BUS_ENTRY_END )
+            || ( IsBus() && item.GetType() != WIRE_END && item.GetType() != PIN_END ) )
         {
-            if( item.GetItem() == this )
-                continue;
-
-            if( ( IsWire() && item.GetType() != BUS_END && item.GetType() != BUS_ENTRY_END )
-                || ( IsBus() && item.GetType() != WIRE_END && item.GetType() != PIN_END ) )
-            {
-                if( m_start == item.GetPosition() )
-                    m_startIsDangling = false;
-
-                if( m_end == item.GetPosition() )
-                    m_endIsDangling = false;
-
-                if( !m_startIsDangling && !m_endIsDangling )
-                    break;
-            }
+            m_startIsDangling = false;
+            break;
         }
-
-        // We only use the bus dangling state for automatic line starting, so we don't care if it
-        // has changed or not (and returning true will result in extra work)
-        if( IsBus() )
-            return false;
-
-        return previousStartState != m_startIsDangling || previousEndState != m_endIsDangling;
     }
 
-    return false;
+    for( auto it = DANGLING_END_ITEM_HELPER::get_lower_pos( aItemListByPos, m_end );
+         it < aItemListByPos.end() && it->GetPosition() == m_end; it++ )
+    {
+        DANGLING_END_ITEM& item = *it;
+
+        if( item.GetItem() == this )
+            continue;
+
+        if( ( IsWire() && item.GetType() != BUS_END && item.GetType() != BUS_ENTRY_END )
+            || ( IsBus() && item.GetType() != WIRE_END && item.GetType() != PIN_END ) )
+        {
+            m_endIsDangling = false;
+            break;
+        }
+    }
+
+    // We only use the bus dangling state for automatic line starting, so we don't care if it
+    // has changed or not (and returning true will result in extra work)
+    if( IsBus() )
+        return false;
+
+    return previousStartState != m_startIsDangling || previousEndState != m_endIsDangling;
 }
 
 
@@ -700,6 +643,7 @@ bool SCH_LINE::CanConnect( const SCH_ITEM* aItem ) const
         case SCH_LABEL_T:
         case SCH_GLOBAL_LABEL_T:
         case SCH_HIER_LABEL_T:
+        case SCH_DIRECTIVE_LABEL_T:
         case SCH_BUS_WIRE_ENTRY_T:
         case SCH_SYMBOL_T:
         case SCH_SHEET_T:
@@ -717,6 +661,7 @@ bool SCH_LINE::CanConnect( const SCH_ITEM* aItem ) const
         case SCH_LABEL_T:
         case SCH_GLOBAL_LABEL_T:
         case SCH_HIER_LABEL_T:
+        case SCH_DIRECTIVE_LABEL_T:
         case SCH_BUS_WIRE_ENTRY_T:
         case SCH_SHEET_T:
         case SCH_SHEET_PIN_T:
@@ -730,7 +675,26 @@ bool SCH_LINE::CanConnect( const SCH_ITEM* aItem ) const
 }
 
 
-std::vector<wxPoint> SCH_LINE::GetConnectionPoints() const
+bool SCH_LINE::HasConnectivityChanges( const SCH_ITEM* aItem,
+                                       const SCH_SHEET_PATH* aInstance ) const
+{
+    // Do not compare to ourself.
+    if( aItem == this || !IsConnectable() )
+        return false;
+
+    const SCH_LINE* line = dynamic_cast<const SCH_LINE*>( aItem );
+
+    // Don't compare against a different SCH_ITEM.
+    wxCHECK( line, false );
+
+    if( GetStartPoint() != line->GetStartPoint() )
+        return true;
+
+    return GetEndPoint() != line->GetEndPoint();
+}
+
+
+std::vector<VECTOR2I> SCH_LINE::GetConnectionPoints() const
 {
     return { m_start, m_end };
 }
@@ -749,7 +713,7 @@ bool SCH_LINE::ConnectionPropagatesTo( const EDA_ITEM* aItem ) const
 }
 
 
-void SCH_LINE::GetSelectedPoints( std::vector< wxPoint >& aPoints ) const
+void SCH_LINE::GetSelectedPoints( std::vector<VECTOR2I>& aPoints ) const
 {
     if( m_flags & STARTPOINT )
         aPoints.push_back( m_start );
@@ -759,9 +723,9 @@ void SCH_LINE::GetSelectedPoints( std::vector< wxPoint >& aPoints ) const
 }
 
 
-wxString SCH_LINE::GetSelectMenuText( EDA_UNITS aUnits ) const
+wxString SCH_LINE::GetItemDescription( UNITS_PROVIDER* aUnitsProvider ) const
 {
-    wxString txtfmt, orient;
+    wxString txtfmt;
 
     if( m_start.x == m_end.x )
     {
@@ -792,7 +756,7 @@ wxString SCH_LINE::GetSelectMenuText( EDA_UNITS aUnits ) const
     }
 
     return wxString::Format( txtfmt,
-                             MessageTextFromValue( aUnits, EuclideanNorm( m_start - m_end ) ) );
+                             aUnitsProvider->MessageTextFromValue( EuclideanNorm( m_start - m_end ) ) );
 }
 
 
@@ -830,7 +794,7 @@ bool SCH_LINE::operator <( const SCH_ITEM& aItem ) const
 }
 
 
-bool SCH_LINE::HitTest( const wxPoint& aPosition, int aAccuracy ) const
+bool SCH_LINE::HitTest( const VECTOR2I& aPosition, int aAccuracy ) const
 {
     // Performance enhancement for connection-building
     if( aPosition == m_start || aPosition == m_end )
@@ -841,22 +805,16 @@ bool SCH_LINE::HitTest( const wxPoint& aPosition, int aAccuracy ) const
     else
         aAccuracy = abs( aAccuracy );
 
-    if( TestSegmentHit( aPosition, m_start, m_end, aAccuracy ) )
-        return true;
-
-    aAccuracy += Mils2iu( DANGLING_SYMBOL_SIZE );
-
-    return ( EuclideanNorm( aPosition - m_start ) < aAccuracy
-            || EuclideanNorm( aPosition - m_end ) < aAccuracy );
+    return TestSegmentHit( aPosition, m_start, m_end, aAccuracy );
 }
 
 
-bool SCH_LINE::HitTest( const EDA_RECT& aRect, bool aContained, int aAccuracy ) const
+bool SCH_LINE::HitTest( const BOX2I& aRect, bool aContained, int aAccuracy ) const
 {
     if( m_flags & (STRUCT_DELETED | SKIP_STRUCT ) )
         return false;
 
-    EDA_RECT rect = aRect;
+    BOX2I rect = aRect;
 
     if ( aAccuracy )
         rect.Inflate( aAccuracy );
@@ -870,6 +828,8 @@ bool SCH_LINE::HitTest( const EDA_RECT& aRect, bool aContained, int aAccuracy )
 
 void SCH_LINE::SwapData( SCH_ITEM* aItem )
 {
+    SCH_ITEM::SwapFlags( aItem );
+
     SCH_LINE* item = (SCH_LINE*) aItem;
 
     std::swap( m_layer, item->m_layer );
@@ -882,7 +842,7 @@ void SCH_LINE::SwapData( SCH_ITEM* aItem )
 }
 
 
-bool SCH_LINE::doIsConnected( const wxPoint& aPosition ) const
+bool SCH_LINE::doIsConnected( const VECTOR2I& aPosition ) const
 {
     if( m_layer != LAYER_WIRE && m_layer != LAYER_BUS )
         return false;
@@ -891,8 +851,12 @@ bool SCH_LINE::doIsConnected( const wxPoint& aPosition ) const
 }
 
 
-void SCH_LINE::Plot( PLOTTER* aPlotter ) const
+void SCH_LINE::Plot( PLOTTER* aPlotter, bool aBackground,
+                     const SCH_PLOT_SETTINGS& aPlotSettings ) const
 {
+    if( aBackground )
+        return;
+
     auto*   settings = static_cast<KIGFX::SCH_RENDER_SETTINGS*>( aPlotter->RenderSettings() );
     int     penWidth = std::max( GetPenWidth(), settings->GetMinPenWidth() );
     COLOR4D color = GetLineColor();
@@ -903,16 +867,50 @@ void SCH_LINE::Plot( PLOTTER* aPlotter ) const
     aPlotter->SetColor( color );
 
     aPlotter->SetCurrentLineWidth( penWidth );
-    aPlotter->SetDash( GetEffectiveLineStyle() );
+    aPlotter->SetDash( penWidth, GetEffectiveLineStyle() );
 
     aPlotter->MoveTo( m_start );
     aPlotter->FinishTo( m_end );
 
-    aPlotter->SetDash( PLOT_DASH_TYPE::SOLID );
+    aPlotter->SetDash( penWidth, LINE_STYLE::SOLID );
+
+    // Plot attributes to a hypertext menu
+    std::vector<wxString> properties;
+    BOX2I                 bbox = GetBoundingBox();
+    bbox.Inflate( GetPenWidth() * 3 );
+
+    if( aPlotSettings.m_PDFPropertyPopups )
+    {
+        if( GetLayer() == LAYER_WIRE )
+        {
+            if( SCH_CONNECTION* connection = Connection() )
+            {
+                properties.emplace_back( wxString::Format( wxT( "!%s = %s" ),
+                                                           _( "Net" ),
+                                                           connection->Name() ) );
+
+                properties.emplace_back( wxString::Format( wxT( "!%s = %s" ),
+                                                           _( "Resolved netclass" ),
+                                                           GetEffectiveNetClass()->GetName() ) );
+            }
+        }
+        else if( GetLayer() == LAYER_BUS )
+        {
+            if( SCH_CONNECTION* connection = Connection() )
+            {
+                for( const std::shared_ptr<SCH_CONNECTION>& member : connection->Members() )
+                    properties.emplace_back( wxT( "!" ) + member->Name() );
+            }
+
+        }
+
+        if( !properties.empty() )
+            aPlotter->HyperlinkMenu( bbox, properties );
+    }
 }
 
 
-void SCH_LINE::SetPosition( const wxPoint& aPosition )
+void SCH_LINE::SetPosition( const VECTOR2I& aPosition )
 {
     m_end = m_end - ( m_start - aPosition );
     m_start = aPosition;
@@ -932,12 +930,12 @@ void SCH_LINE::GetMsgPanelInfo( EDA_DRAW_FRAME* aFrame, std::vector<MSG_PANEL_IT
 
     aList.emplace_back( _( "Line Type" ), msg );
 
-    if( GetLineStyle() != GetEffectiveLineStyle() )
-        msg = _( "from netclass" );
-    else
-        msg = GetLineStyleName( GetLineStyle() );
+    LINE_STYLE lineStyle = GetLineStyle();
 
-    aList.emplace_back( _( "Line Style" ), msg );
+    if( GetEffectiveLineStyle() != lineStyle )
+        aList.emplace_back( _( "Line Style" ), _( "from netclass" ) );
+    else
+        m_stroke.GetMsgPanelInfo( aFrame, aList, true, false );
 
     SCH_CONNECTION* conn = nullptr;
 
@@ -950,14 +948,8 @@ void SCH_LINE::GetMsgPanelInfo( EDA_DRAW_FRAME* aFrame, std::vector<MSG_PANEL_IT
 
         if( !conn->IsBus() )
         {
-            NET_SETTINGS& netSettings = Schematic()->Prj().GetProjectFile().NetSettings();
-            wxString netname = conn->Name();
-            wxString netclassName = netSettings.m_NetClasses.GetDefaultPtr()->GetName();
-
-            if( netSettings.m_NetClassAssignments.count( netname ) )
-                netclassName = netSettings.m_NetClassAssignments[ netname ];
-
-            aList.emplace_back( _( "Assigned Netclass" ), netclassName );
+            aList.emplace_back( _( "Resolved Netclass" ),
+                                UnescapeString( GetEffectiveNetClass()->GetName() ) );
         }
     }
 }
@@ -974,15 +966,105 @@ bool SCH_LINE::IsWire() const
     return ( GetLayer() == LAYER_WIRE );
 }
 
+
 bool SCH_LINE::IsBus() const
 {
     return ( GetLayer() == LAYER_BUS );
 }
 
 
-bool SCH_LINE::UsesDefaultStroke() const
+bool SCH_LINE::operator==( const SCH_ITEM& aOther ) const
 {
-    return m_stroke.GetWidth() == 0 && m_stroke.GetColor() == COLOR4D::UNSPECIFIED
-            && ( m_stroke.GetPlotStyle() == GetDefaultStyle()
-            || m_stroke.GetPlotStyle() == PLOT_DASH_TYPE::DEFAULT );
+    if( Type() != aOther.Type() )
+        return false;
+
+    const SCH_LINE& other = static_cast<const SCH_LINE&>( aOther );
+
+    if( GetLayer() != other.GetLayer() )
+        return false;
+
+    if( m_start != other.m_start )
+        return false;
+
+    if( m_end != other.m_end )
+        return false;
+
+    if( m_stroke.GetWidth() != other.m_stroke.GetWidth() )
+        return false;
+
+    if( m_stroke.GetColor() != other.m_stroke.GetColor() )
+        return false;
+
+    if( m_stroke.GetLineStyle() != other.m_stroke.GetLineStyle() )
+        return false;
+
+    return true;
 }
+
+
+double SCH_LINE::Similarity( const SCH_ITEM& aOther ) const
+{
+    if( m_Uuid == aOther.m_Uuid )
+        return 1.0;
+
+    if( Type() != aOther.Type() )
+        return 0.0;
+
+    const SCH_LINE& other = static_cast<const SCH_LINE&>( aOther );
+
+    if( GetLayer() != other.GetLayer() )
+        return 0.0;
+
+    double similarity = 1.0;
+
+    if( m_start != other.m_start )
+        similarity *= 0.9;
+
+    if( m_end != other.m_end )
+        similarity *= 0.9;
+
+    if( m_stroke.GetWidth() != other.m_stroke.GetWidth() )
+        similarity *= 0.9;
+
+    if( m_stroke.GetColor() != other.m_stroke.GetColor() )
+        similarity *= 0.9;
+
+    if( m_stroke.GetLineStyle() != other.m_stroke.GetLineStyle() )
+        similarity *= 0.9;
+
+    return similarity;
+}
+
+
+static struct SCH_LINE_DESC
+{
+    SCH_LINE_DESC()
+    {
+        ENUM_MAP<LINE_STYLE>& plotDashTypeEnum = ENUM_MAP<LINE_STYLE>::Instance();
+
+        if( plotDashTypeEnum.Choices().GetCount() == 0 )
+        {
+            plotDashTypeEnum.Map( LINE_STYLE::DEFAULT, _HKI( "Default" ) )
+                            .Map( LINE_STYLE::SOLID, _HKI( "Solid" ) )
+                            .Map( LINE_STYLE::DASH, _HKI( "Dashed" ) )
+                            .Map( LINE_STYLE::DOT, _HKI( "Dotted" ) )
+                            .Map( LINE_STYLE::DASHDOT, _HKI( "Dash-Dot" ) )
+                            .Map( LINE_STYLE::DASHDOTDOT, _HKI( "Dash-Dot-Dot" ) );
+        }
+
+        PROPERTY_MANAGER& propMgr = PROPERTY_MANAGER::Instance();
+        REGISTER_TYPE( SCH_LINE );
+        propMgr.InheritsAfter( TYPE_HASH( SCH_LINE ), TYPE_HASH( SCH_ITEM ) );
+
+        void ( SCH_LINE::*lineStyleSetter )( LINE_STYLE ) = &SCH_LINE::SetLineStyle;
+
+        propMgr.AddProperty( new PROPERTY_ENUM<SCH_LINE, LINE_STYLE>( _HKI( "Line Style" ),
+                lineStyleSetter, &SCH_LINE::GetLineStyle ) );
+
+        propMgr.AddProperty( new PROPERTY<SCH_LINE, int>( _HKI( "Line Width" ),
+                &SCH_LINE::SetLineWidth, &SCH_LINE::GetLineWidth, PROPERTY_DISPLAY::PT_SIZE ) );
+
+        propMgr.AddProperty( new PROPERTY<SCH_LINE, COLOR4D>( _HKI( "Color" ),
+                &SCH_LINE::SetLineColor, &SCH_LINE::GetLineColor ) );
+    }
+} _SCH_LINE_DESC;
