@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2009 Isaac Marino Bavaresco, isaacbavaresco@yahoo.com.br
  * Copyright (C) 2009 SoftPLC Corporation, Dick Hollenbeck <dick@softplc.com>
- * Copyright (C) 2009-2021 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright (C) 2009-2024 KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -27,15 +27,18 @@
 #include <confirm.h>
 #include <core/arraydim.h>
 #include <core/kicad_algo.h>
-#include <pcbnew.h>
 #include <pcb_edit_frame.h>
+#include <tool/tool_manager.h>
+#include <tools/pcb_actions.h>
 #include <board.h>
 #include <collectors.h>
+#include <footprint.h>
+#include <pad.h>
+#include <pcb_track.h>
 #include <panel_setup_layers.h>
 #include <board_stackup_manager/panel_board_stackup.h>
 
 #include <wx/choicdlg.h>
-#include <wx/treebook.h>
 #include <eda_list_dialog.h>
 
 
@@ -49,7 +52,7 @@
 static LSEQ dlg_layers()
 {
     // Layers that are put out into the dialog UI, coordinate with wxformbuilder and
-    // getCTLs( LAYER_NUM aLayerNumber )
+    // getCTLs( int aLayerNumber )
     static const PCB_LAYER_ID layers[] = {
         F_CrtYd,
         F_Fab,
@@ -121,17 +124,17 @@ static LSEQ dlg_layers()
 }
 
 
-PANEL_SETUP_LAYERS::PANEL_SETUP_LAYERS( PAGED_DIALOG* aParent, PCB_EDIT_FRAME* aFrame ) :
-        PANEL_SETUP_LAYERS_BASE( aParent->GetTreebook() ),
-        m_parentDialog( aParent ),
+PANEL_SETUP_LAYERS::PANEL_SETUP_LAYERS( wxWindow* aParentWindow, PCB_EDIT_FRAME* aFrame ) :
+        PANEL_SETUP_LAYERS_BASE( aParentWindow ),
         m_frame( aFrame ),
-        m_physicalStackup( nullptr )
+        m_physicalStackup( nullptr ),
+        m_initialized( false )
 {
     m_pcb = aFrame->GetBoard();
 }
 
 
-PANEL_SETUP_LAYERS_CTLs PANEL_SETUP_LAYERS::getCTLs( LAYER_NUM aLayerNumber )
+PANEL_SETUP_LAYERS_CTLs PANEL_SETUP_LAYERS::getCTLs( int aLayerNumber )
 {
 #define RETURN_COPPER( x )    return PANEL_SETUP_LAYERS_CTLs( x##Name, x##CheckBox, x##Choice )
 #define RETURN_AUX( x )       return PANEL_SETUP_LAYERS_CTLs( x##Name, x##CheckBox, x##StaticText )
@@ -139,70 +142,70 @@ PANEL_SETUP_LAYERS_CTLs PANEL_SETUP_LAYERS::getCTLs( LAYER_NUM aLayerNumber )
 
     switch( aLayerNumber )
     {
-    case F_CrtYd:               RETURN_MANDATORY( m_CrtYdFront );
-    case F_Fab:                 RETURN_AUX( m_FabFront );
-    case F_Adhes:               RETURN_AUX( m_AdhesFront );
-    case F_Paste:               RETURN_AUX( m_SoldPFront );
-    case F_SilkS:               RETURN_AUX( m_SilkSFront );
-    case F_Mask:                RETURN_AUX( m_MaskFront );
-    case F_Cu:                  RETURN_COPPER( m_Front );
+    case F_CrtYd:      RETURN_MANDATORY( m_CrtYdFront );
+    case F_Fab:        RETURN_AUX( m_FabFront );
+    case F_Adhes:      RETURN_AUX( m_AdhesFront );
+    case F_Paste:      RETURN_AUX( m_SoldPFront );
+    case F_SilkS:      RETURN_AUX( m_SilkSFront );
+    case F_Mask:       RETURN_AUX( m_MaskFront );
+    case F_Cu:         RETURN_COPPER( m_Front );
 
-    case In1_Cu:                RETURN_COPPER( m_In1 );
-    case In2_Cu:                RETURN_COPPER( m_In2 );
-    case In3_Cu:                RETURN_COPPER( m_In3 );
-    case In4_Cu:                RETURN_COPPER( m_In4 );
-    case In5_Cu:                RETURN_COPPER( m_In5 );
-    case In6_Cu:                RETURN_COPPER( m_In6 );
-    case In7_Cu:                RETURN_COPPER( m_In7 );
-    case In8_Cu:                RETURN_COPPER( m_In8 );
-    case In9_Cu:                RETURN_COPPER( m_In9 );
-    case In10_Cu:               RETURN_COPPER( m_In10 );
-    case In11_Cu:               RETURN_COPPER( m_In11 );
-    case In12_Cu:               RETURN_COPPER( m_In12 );
-    case In13_Cu:               RETURN_COPPER( m_In13 );
-    case In14_Cu:               RETURN_COPPER( m_In14 );
-    case In15_Cu:               RETURN_COPPER( m_In15 );
+    case In1_Cu:       RETURN_COPPER( m_In1 );
+    case In2_Cu:       RETURN_COPPER( m_In2 );
+    case In3_Cu:       RETURN_COPPER( m_In3 );
+    case In4_Cu:       RETURN_COPPER( m_In4 );
+    case In5_Cu:       RETURN_COPPER( m_In5 );
+    case In6_Cu:       RETURN_COPPER( m_In6 );
+    case In7_Cu:       RETURN_COPPER( m_In7 );
+    case In8_Cu:       RETURN_COPPER( m_In8 );
+    case In9_Cu:       RETURN_COPPER( m_In9 );
+    case In10_Cu:      RETURN_COPPER( m_In10 );
+    case In11_Cu:      RETURN_COPPER( m_In11 );
+    case In12_Cu:      RETURN_COPPER( m_In12 );
+    case In13_Cu:      RETURN_COPPER( m_In13 );
+    case In14_Cu:      RETURN_COPPER( m_In14 );
+    case In15_Cu:      RETURN_COPPER( m_In15 );
 
-    case In16_Cu:               RETURN_COPPER( m_In16 );
-    case In17_Cu:               RETURN_COPPER( m_In17 );
-    case In18_Cu:               RETURN_COPPER( m_In18 );
-    case In19_Cu:               RETURN_COPPER( m_In19 );
-    case In20_Cu:               RETURN_COPPER( m_In20 );
-    case In21_Cu:               RETURN_COPPER( m_In21 );
-    case In22_Cu:               RETURN_COPPER( m_In22 );
-    case In23_Cu:               RETURN_COPPER( m_In23 );
-    case In24_Cu:               RETURN_COPPER( m_In24 );
-    case In25_Cu:               RETURN_COPPER( m_In25 );
-    case In26_Cu:               RETURN_COPPER( m_In26 );
-    case In27_Cu:               RETURN_COPPER( m_In27 );
-    case In28_Cu:               RETURN_COPPER( m_In28 );
-    case In29_Cu:               RETURN_COPPER( m_In29 );
-    case In30_Cu:               RETURN_COPPER( m_In30 );
+    case In16_Cu:      RETURN_COPPER( m_In16 );
+    case In17_Cu:      RETURN_COPPER( m_In17 );
+    case In18_Cu:      RETURN_COPPER( m_In18 );
+    case In19_Cu:      RETURN_COPPER( m_In19 );
+    case In20_Cu:      RETURN_COPPER( m_In20 );
+    case In21_Cu:      RETURN_COPPER( m_In21 );
+    case In22_Cu:      RETURN_COPPER( m_In22 );
+    case In23_Cu:      RETURN_COPPER( m_In23 );
+    case In24_Cu:      RETURN_COPPER( m_In24 );
+    case In25_Cu:      RETURN_COPPER( m_In25 );
+    case In26_Cu:      RETURN_COPPER( m_In26 );
+    case In27_Cu:      RETURN_COPPER( m_In27 );
+    case In28_Cu:      RETURN_COPPER( m_In28 );
+    case In29_Cu:      RETURN_COPPER( m_In29 );
+    case In30_Cu:      RETURN_COPPER( m_In30 );
 
-    case B_Cu:                  RETURN_COPPER( m_Back );
-    case B_Mask:                RETURN_AUX( m_MaskBack );
-    case B_SilkS:               RETURN_AUX( m_SilkSBack );
-    case B_Paste:               RETURN_AUX( m_SoldPBack );
-    case B_Adhes:               RETURN_AUX( m_AdhesBack );
-    case B_Fab:                 RETURN_AUX( m_FabBack );
-    case B_CrtYd:               RETURN_MANDATORY( m_CrtYdBack );
+    case B_Cu:         RETURN_COPPER( m_Back );
+    case B_Mask:       RETURN_AUX( m_MaskBack );
+    case B_SilkS:      RETURN_AUX( m_SilkSBack );
+    case B_Paste:      RETURN_AUX( m_SoldPBack );
+    case B_Adhes:      RETURN_AUX( m_AdhesBack );
+    case B_Fab:        RETURN_AUX( m_FabBack );
+    case B_CrtYd:      RETURN_MANDATORY( m_CrtYdBack );
 
-    case Edge_Cuts:             RETURN_MANDATORY( m_PCBEdges );
-    case Margin:                RETURN_MANDATORY( m_Margin );
-    case Eco2_User:             RETURN_AUX( m_Eco2 );
-    case Eco1_User:             RETURN_AUX( m_Eco1 );
-    case Cmts_User:             RETURN_AUX( m_Comments );
-    case Dwgs_User:             RETURN_AUX( m_Drawings );
+    case Edge_Cuts:    RETURN_MANDATORY( m_PCBEdges );
+    case Margin:       RETURN_MANDATORY( m_Margin );
+    case Eco2_User:    RETURN_AUX( m_Eco2 );
+    case Eco1_User:    RETURN_AUX( m_Eco1 );
+    case Cmts_User:    RETURN_AUX( m_Comments );
+    case Dwgs_User:    RETURN_AUX( m_Drawings );
 
-    case User_1:                RETURN_AUX( m_User1 );
-    case User_2:                RETURN_AUX( m_User2 );
-    case User_3:                RETURN_AUX( m_User3 );
-    case User_4:                RETURN_AUX( m_User4 );
-    case User_5:                RETURN_AUX( m_User5 );
-    case User_6:                RETURN_AUX( m_User6 );
-    case User_7:                RETURN_AUX( m_User7 );
-    case User_8:                RETURN_AUX( m_User8 );
-    case User_9:                RETURN_AUX( m_User9 );
+    case User_1:       RETURN_AUX( m_User1 );
+    case User_2:       RETURN_AUX( m_User2 );
+    case User_3:       RETURN_AUX( m_User3 );
+    case User_4:       RETURN_AUX( m_User4 );
+    case User_5:       RETURN_AUX( m_User5 );
+    case User_6:       RETURN_AUX( m_User6 );
+    case User_7:       RETURN_AUX( m_User7 );
+    case User_8:       RETURN_AUX( m_User8 );
+    case User_9:       RETURN_AUX( m_User9 );
 
     default:
         wxASSERT_MSG( 0, wxT( "bad layer id" ) );
@@ -211,19 +214,19 @@ PANEL_SETUP_LAYERS_CTLs PANEL_SETUP_LAYERS::getCTLs( LAYER_NUM aLayerNumber )
 }
 
 
-wxControl* PANEL_SETUP_LAYERS::getName( LAYER_NUM aLayer )
+wxControl* PANEL_SETUP_LAYERS::getName( int aLayer )
 {
     return getCTLs( aLayer ).name;
 }
 
 
-wxCheckBox* PANEL_SETUP_LAYERS::getCheckBox( LAYER_NUM aLayer )
+wxCheckBox* PANEL_SETUP_LAYERS::getCheckBox( int aLayer )
 {
     return getCTLs( aLayer ).checkbox;
 }
 
 
-wxChoice* PANEL_SETUP_LAYERS::getChoice( LAYER_NUM aLayer )
+wxChoice* PANEL_SETUP_LAYERS::getChoice( int aLayer )
 {
     return (wxChoice*) getCTLs( aLayer ).choice;
 }
@@ -245,6 +248,8 @@ bool PANEL_SETUP_LAYERS::TransferDataToWindow()
     setMandatoryLayerCheckBoxes();
     setUserDefinedLayerCheckBoxes();
 
+    m_initialized = true;
+
     return true;
 }
 
@@ -267,7 +272,7 @@ void PANEL_SETUP_LAYERS::setUserDefinedLayerCheckBoxes()
     for( LSEQ seq = LSET::UserDefinedLayers().Seq();  seq;  ++seq )
     {
         PCB_LAYER_ID layer = *seq;
-        bool     state = m_pcb->IsLayerEnabled( layer );
+        bool         state = m_pcb->IsLayerEnabled( layer );
 
 #ifdef HIDE_INACTIVE_LAYERS
         // This code hides inactive copper layers, or redisplays hidden layers which are now needed.
@@ -308,7 +313,7 @@ void PANEL_SETUP_LAYERS::showBoardLayerNames()
             wxString lname = m_pcb->GetLayerName( layer );
 
             if( auto textCtl = dynamic_cast<wxTextCtrl*>( ctl ) )
-                textCtl->SetValue( lname );     // wxTextCtrl
+                textCtl->ChangeValue( lname );     // wxTextCtrl
             else
                 ctl->SetLabel( lname );         // wxStaticText
         }
@@ -356,7 +361,7 @@ LSET PANEL_SETUP_LAYERS::GetUILayerMask()
 }
 
 
-void PANEL_SETUP_LAYERS::setLayerCheckBox( LAYER_NUM aLayer, bool isChecked )
+void PANEL_SETUP_LAYERS::setLayerCheckBox( int aLayer, bool isChecked )
 {
     PANEL_SETUP_LAYERS_CTLs ctl = getCTLs( aLayer );
 
@@ -419,7 +424,6 @@ void PANEL_SETUP_LAYERS::OnCheckBox( wxCommandEvent& event )
 void PANEL_SETUP_LAYERS::DenyChangeCheckBox( wxCommandEvent& event )
 {
     wxObject* source = event.GetEventObject();
-    wxString msg;
 
     for( LSEQ seq = LSET::AllCuMask().Seq(); seq; ++seq )
     {
@@ -427,8 +431,9 @@ void PANEL_SETUP_LAYERS::DenyChangeCheckBox( wxCommandEvent& event )
 
         if( source == copper )
         {
-            DisplayError( this,
-                    _( "Use the Physical Stackup page to change the number of copper layers." ) );
+            DisplayError( wxGetTopLevelParent( this ),
+                          _( "Use the Physical Stackup page to change the number of "
+                             "copper layers." ) );
 
             copper->SetValue( true );
             return;
@@ -442,10 +447,9 @@ bool PANEL_SETUP_LAYERS::TransferDataFromWindow()
     if( !testLayerNames() )
         return false;
 
-    wxASSERT( m_physicalStackup );
-
     // Make sure we have the latest copper layer count
-    SyncCopperLayers( m_physicalStackup->GetCopperLayerCount() );
+    if( m_physicalStackup )
+        SyncCopperLayers( m_physicalStackup->GetCopperLayerCount() );
 
     wxString msg;
     bool     modified = false;
@@ -459,21 +463,23 @@ bool PANEL_SETUP_LAYERS::TransferDataFromWindow()
     if( !notremovableLayers.empty() )
     {
         for( unsigned int ii = 0; ii < notremovableLayers.size(); ii++ )
-            msg << m_pcb->GetLayerName( notremovableLayers[ii] ) << "\n";
+            msg << m_pcb->GetLayerName( notremovableLayers[ii] ) << wxT( "\n" );
 
-        if( !IsOK( this, wxString::Format( _( "Footprints have some items on removed layers:\n"
-                                              "%s\n"
-                                              "These items will be no longer accessible\n"
-                                              "Do you wish to continue?" ), msg ) ) )
+        if( !IsOK( wxGetTopLevelParent( this ),
+                   wxString::Format( _( "Footprints have some items on removed layers:\n"
+                                        "%s\n"
+                                        "These items will be no longer accessible\n"
+                                        "Do you wish to continue?" ), msg ) ) )
         {
             return false;
         }
     }
 
     if( !removedLayers.empty()
-            && !IsOK( this, _( "Items have been found on removed layers. This operation will "
-                               "delete all items from removed layers and cannot be undone.\n"
-                               "Do you wish to continue?" ) ) )
+      && !IsOK( wxGetTopLevelParent( this ),
+                _( "Items have been found on removed layers. This operation will "
+                   "delete all items from removed layers and cannot be undone.\n"
+                   "Do you wish to continue?" ) ) )
     {
         return false;
     }
@@ -484,6 +490,8 @@ bool PANEL_SETUP_LAYERS::TransferDataFromWindow()
 
     if( !removedLayers.empty() )
     {
+        m_frame->GetToolManager()->RunAction( PCB_ACTIONS::selectionClear );
+
         PCB_LAYER_COLLECTOR collector;
 
         for( PCB_LAYER_ID layer_id : removedLayers )
@@ -495,36 +503,98 @@ bool PANEL_SETUP_LAYERS::TransferDataFromWindow()
             for( int i = 0; i < collector.GetCount(); i++ )
             {
                 BOARD_ITEM* item = collector[i];
-                LSET        layers = item->GetLayerSet();
 
-                layers.reset( layer_id );
-                hasRemovedBoardItems = true;
-                modified = true;
+                // Do not remove/change an item owned by a footprint
+                if( item->GetParentFootprint() )
+                    continue;
 
-                if( layers.any() )
-                {
-                    item->SetLayerSet( layers );
-                }
-                else
+                // Do not remove footprints
+                if( item->Type() == PCB_FOOTPRINT_T )
+                    continue;
+
+                // Vias are also specific
+                // Note: vias are specific. They are only on copper layers,  and
+                // do not use a layer set, only store the copper top and the copper bottom.
+                // So reinit the layer set does not work with vias
+                if( item->Type() == PCB_VIA_T )
                 {
+                    PCB_VIA* via = static_cast<PCB_VIA*>( item );
+
+                    if( via->GetViaType() == VIATYPE::THROUGH )
+                        continue;
+                    else
+                    {
+                        PCB_LAYER_ID top_layer;
+                        PCB_LAYER_ID bottom_layer;
+                        via->LayerPair( &top_layer, &bottom_layer );
+
+                        if( top_layer != layer_id && bottom_layer != layer_id )
+                            continue;
+                    }
+                    // blind/buried vias with a top or bottom layer on a removed layer
+                    // are removed. Perhaps one could just modify the top/bottom layer,
+                    // but I am not sure this is better.
                     m_pcb->Remove( item );
                     delete item;
                 }
+                else
+                {
+                    LSET        layers = item->GetLayerSet();
+
+                    layers.reset( layer_id );
+                    hasRemovedBoardItems = true;
+                    modified = true;
+
+                    if( layers.any() )
+                    {
+                        item->SetLayerSet( layers );
+                    }
+                    else
+                    {
+                        m_pcb->Remove( item );
+                        delete item;
+                    }
+                }
             }
         }
+
+        // Undo state may have copies of pointers deleted above
+        m_frame->ClearUndoRedoList();
     }
 
     m_enabledLayers = GetUILayerMask();
 
-    if( m_enabledLayers != m_pcb->GetEnabledLayers() )
+    LSET previousEnabled = m_pcb->GetEnabledLayers();
+
+    if( m_enabledLayers != previousEnabled )
     {
         m_pcb->SetEnabledLayers( m_enabledLayers );
 
-        /* Ensure enabled layers are also visible
-         * This is mainly to avoid mistakes if some enabled
-         * layers are not visible when exiting this dialog
+        LSET changedLayers = m_enabledLayers ^ previousEnabled;
+
+        /*
+         * Ensure enabled layers are also visible.  This is mainly to avoid mistakes if some
+         * enabled layers are not visible when exiting this dialog.
          */
-        m_pcb->SetVisibleLayers( m_enabledLayers );
+        m_pcb->SetVisibleLayers( m_pcb->GetVisibleLayers() | changedLayers );
+
+        /*
+         * Ensure items with through holes have all inner copper layers.  (For historical reasons
+         * this is NOT trimmed to the currently-enabled inner layers.)
+         */
+        for( FOOTPRINT* fp : m_pcb->Footprints() )
+        {
+            for( PAD* pad : fp->Pads() )
+            {
+                if( pad->HasHole() && pad->IsOnCopperLayer() )
+                    pad->SetLayerSet( pad->GetLayerSet() | LSET::InternalCuMask() );
+            }
+        }
+
+        // Tracks do not change their layer
+        // Vias layers are defined by the starting layer and the ending layer, so
+        // they are not modified by adding a layer.
+        // So do nothing for tracks/vias
 
         modified = true;
     }
@@ -569,14 +639,10 @@ bool PANEL_SETUP_LAYERS::TransferDataFromWindow()
         }
     }
 
-    // If some board items are deleted: Rebuild the connectivity,
-    // because it is likely some tracks and vias were removed
+    // If some board items are deleted: Rebuild the connectivity, because it is likely some
+    // tracks and vias were removed
     if( hasRemovedBoardItems )
-    {
-        // Rebuild list of nets (full ratsnest rebuild)
         m_pcb->BuildConnectivity();
-        m_frame->Compile_Ratsnest( true );
-    }
 
     if( modified )
         m_frame->OnModify();
@@ -585,7 +651,7 @@ bool PANEL_SETUP_LAYERS::TransferDataFromWindow()
 }
 
 
-int PANEL_SETUP_LAYERS::getLayerTypeIndex( LAYER_NUM aLayer )
+int PANEL_SETUP_LAYERS::getLayerTypeIndex( int aLayer )
 {
     wxChoice* ctl = getChoice( aLayer );
     int ret = ctl->GetCurrentSelection(); // Indices must have same sequence as LAYER_T
@@ -593,7 +659,7 @@ int PANEL_SETUP_LAYERS::getLayerTypeIndex( LAYER_NUM aLayer )
 }
 
 
-wxString PANEL_SETUP_LAYERS::GetLayerName( LAYER_NUM aLayer )
+wxString PANEL_SETUP_LAYERS::GetLayerName( int aLayer )
 {
     wxControl* control = getName( aLayer );
 
@@ -647,20 +713,21 @@ bool PANEL_SETUP_LAYERS::testLayerNames()
 
         if( !name )
         {
-            m_parentDialog->SetError( _( "Layer must have a name." ), this, ctl );
+            PAGED_DIALOG::GetDialog( this )->SetError( _( "Layer must have a name." ), this, ctl );
             return false;
         }
 
         if( hasOneOf( name, badchars ) )
         {
             wxString msg = wxString::Format(_( "%s are forbidden in layer names." ), badchars );
-            m_parentDialog->SetError( msg, this, ctl );
+            PAGED_DIALOG::GetDialog( this )->SetError( msg, this, ctl );
             return false;
         }
 
         if( name == wxT( "signal" ) )
         {
-            m_parentDialog->SetError( _( "Layer name \"signal\" is reserved." ), this, ctl );
+            PAGED_DIALOG::GetDialog( this )->SetError( _( "Layer name \"signal\" is reserved." ),
+                                                       this, ctl );
             return false;
         }
 
@@ -669,7 +736,7 @@ bool PANEL_SETUP_LAYERS::testLayerNames()
             if( name == existingName )
             {
                 wxString msg = wxString::Format(_( "Layer name '%s' already in use." ), name );
-                m_parentDialog->SetError( msg, this, ctl );
+                PAGED_DIALOG::GetDialog( this )->SetError( msg, this, ctl );
                 return false;
             }
         }
@@ -701,7 +768,40 @@ LSEQ PANEL_SETUP_LAYERS::getRemovedLayersWithItems()
             collector.Collect( m_pcb, GENERAL_COLLECTOR::BoardLevelItems );
 
             if( collector.GetCount() != 0 )
-                removedLayers.push_back( layer_id );
+            {
+                // Skip items owned by footprints and footprints when building
+                // the actual list of removed layers: these items are not removed
+                for( int i = 0; i < collector.GetCount(); i++ )
+                {
+                    BOARD_ITEM* item = collector[i];
+
+                    if( item->Type() == PCB_FOOTPRINT_T || item->GetParentFootprint() )
+                        continue;
+
+                    // Vias are on multiple adjacent layers, but only the top and
+                    // the bottom layers are stored. So there are issues only if one
+                    // is on a removed layer
+                    if( item->Type() == PCB_VIA_T )
+                    {
+                        PCB_VIA* via = static_cast<PCB_VIA*>( item );
+
+                        if( via->GetViaType() == VIATYPE::THROUGH )
+                            continue;
+                        else
+                        {
+                            PCB_LAYER_ID top_layer;
+                            PCB_LAYER_ID bottom_layer;
+                            via->LayerPair( &top_layer, &bottom_layer );
+
+                            if( top_layer != layer_id && bottom_layer != layer_id )
+                                continue;
+                        }
+                    }
+
+                    removedLayers.push_back( layer_id );
+                    break;
+                }
+            }
         }
     }
 
@@ -722,7 +822,7 @@ LSEQ PANEL_SETUP_LAYERS::getNonRemovableLayers()
     PCB_LAYER_COLLECTOR collector;
     LSEQ newLayerSeq = newLayers.Seq();
 
-    for( auto layer_id : curLayers.Seq() )
+    for( PCB_LAYER_ID layer_id : curLayers.Seq() )
     {
         if( IsCopperLayer( layer_id ) ) // Copper layers are not taken into account here
             continue;
@@ -775,7 +875,9 @@ bool PANEL_SETUP_LAYERS::CheckCopperLayerCount( BOARD* aWorkingBoard, BOARD* aIm
                                          newNumLayers,
                                          currNumLayers );
 
-        wxMessageDialog dlg( this, msg, _( "Inner Layers To Be Deleted" ),
+        wxWindow* topLevelParent = wxGetTopLevelParent( this );
+
+        wxMessageDialog dlg( topLevelParent, msg, _( "Inner Layers To Be Deleted" ),
                              wxICON_WARNING | wxSTAY_ON_TOP | wxYES | wxNO | wxNO_DEFAULT );
 
         if( wxID_NO == dlg.ShowModal() )
@@ -809,12 +911,13 @@ void PANEL_SETUP_LAYERS::addUserDefinedLayer( wxCommandEvent& aEvent )
 
     if( list.empty() )
     {
-        DisplayErrorMessage( m_parentDialog,
+        DisplayErrorMessage( PAGED_DIALOG::GetDialog( this ),
                              _( "All user-defined layers have already been added." ) );
         return;
     }
 
-    EDA_LIST_DIALOG dlg( m_parentDialog, _( "Add User-defined Layer" ), headers, list );
+    EDA_LIST_DIALOG dlg( PAGED_DIALOG::GetDialog( this ), _( "Add User-defined Layer" ),
+                         headers, list );
     dlg.SetListLabel( _( "Select layer to add:" ) );
     dlg.HideFilter();
 
